#include <Wire.h>
#include <hd44780.h>                       // main hd44780 header
#include <hd44780ioClass/hd44780_I2Cexp.h> // i2c expander i/o class header

hd44780_I2Cexp lcd; // declare lcd object: auto locate & config exapander chip

// LCD geometry
const int LCD_COLS = 16;
const int LCD_ROWS = 2;

// minature digits
uint8_t minidigit[10][8] = {
	{0x07,0x05,0x05,0x05,0x07,0x00,0x00,0x00}, // minature 0
	{0x02,0x06,0x02,0x02,0x07,0x00,0x00,0x00}, // minature 1
	{0x03,0x05,0x02,0x04,0x07,0x00,0x00,0x00}, // minature 2
	{0x07,0x01,0x07,0x01,0x07,0x00,0x00,0x00}, // minature 3
	{0x05,0x05,0x07,0x01,0x01,0x00,0x00,0x00}, // minature 4
	{0x07,0x04,0x06,0x01,0x07,0x00,0x00,0x00}, // minature 5
	{0x07,0x04,0x07,0x05,0x07,0x00,0x00,0x00}, // minature 6
	{0x07,0x01,0x02,0x04,0x04,0x00,0x00,0x00}, // minature 7
	{0x07,0x05,0x07,0x05,0x07,0x00,0x00,0x00}, // minature 8
	{0x07,0x05,0x07,0x01,0x01,0x00,0x00,0x00}, // minature 9
};

void setup()
{
int status;

	delay(500);

	Serial.begin(115200);

	status = lcd.begin(LCD_COLS, LCD_ROWS);
	if(status) // non zero status means it was unsuccesful
	{
		hd44780::fatalError(status); // does not return
	}
	// initalization was successful, the backlight should be on now

	Pprintf(Serial, "\nCGRAM at powerup:\n");
	// first dump CGRAM to see what is in it
	for(uint8_t ccode = 0; ccode < 8; ccode++)
	{
		Pprintf(Serial, "Char code: %02X: ", ccode);
		dumpCGRAM(Serial, lcd, ccode); // dump CGRAM for ccode to Serial
	}


	// Print a message to the LCD
	// note: lcd.println() is NOT used since hd44780 errors off this call.
	// so send <CR> and <LF> manually
	Pprintf(lcd, "<CR>:%c <LF>:%c", '\r', '\n');

	// turn this on to see tiny digit for each custom character code point
	// Just be aware that after this runs once, you have to power cycle
	// the LCD to get the defaults back.
	// NOTE: the LCD can maintain enough power for the CGRAM for many	
	// seconds so it is best to actually short VCC and GND on the LCD after
	// power is removed and before reapplying power.

#define CUSTOMCC
#if defined(CUSTOMCC)
	delay(5000);

	// note: once custom characters are defined, the <CR> and <LF> characters
	// will change to their custom character definitions.
	for(uint8_t c = 0; c < 8; c++)
		lcd.createChar(c, minidigit[c]);

	//  dump CGRAM again to see what is in it
	Pprintf(Serial, "\n");
	Pprintf(Serial, "CGRAM after custom chars\n");
	for(uint8_t ccode = 0; ccode < 8; ccode++)
	{
		Pprintf(Serial, "Char code: %02X: ", ccode);
		dumpCGRAM(Serial, lcd, ccode); // dump CGRAM for ccode to Serial
	}
#endif

}

void loop(){}

void dumpCGRAM(Print &outdev, hd44780 &lcd, uint8_t charcode)
{
int CGdata;
int ddramaddr;

	ddramaddr = lcd.status(); // fetch status which includes ddram address

	if(ddramaddr < 0) // status() failed, so just set address to 0
		ddramaddr = 0;
	else
		ddramaddr &= 0x7f; // strip off BUSY bit

	lcd.command(hd44780::HD44780_SETCGRAMADDR | (charcode << 3));

	for(uint8_t n = 0; n < 8; n++)
	{
		CGdata = lcd.read();
		if(CGdata < 0)
		{
			Pprintf(outdev, "read() error\n");
			break;
		}
		else
		{
			Pprintf(outdev, "%02X ", CGdata);
		}
	}
	Pprintf(outdev,"\n");

	// put LCD back into DDRAM mode so write() works
	lcd.setCursor(ddramaddr , 0);
	
}
#ifndef PPRINTF_BUFSIZE
#define PPRINTF_BUFSIZE 64
#endif
size_t Pprintf(Print &outdev, const char *format, ...)
{
char buf[PPRINTF_BUFSIZE];
	va_list ap;
	va_start(ap, format);
	vsnprintf(buf, sizeof(buf), format, ap);
	va_end(ap);
	return(outdev.write(buf));
}
